from gscdash.course_info import Assignment

import requests
from bs4 import BeautifulSoup
from gscdash.pyscope.question import GSQuestion
import json

class GSAssignment(Assignment):

    def __init__(self, name, aid, points, percent_graded, complete, regrades_on, course):
        '''Create a assignment object'''
        self.name = name
        self.aid = aid
        self.assignment_uuid = aid
        self.points = points
        self.percent_graded = percent_graded
        self.complete = complete
        self.regrades_on = regrades_on
        self.course = course
        self.questions = []

    def add_question(self, title, weight, crop = None, content = [], parent_id = None):
        new_q_data = [q.to_patch() for q in self.questions]
        new_crop = crop if crop else [{'x1': 10, 'x2': 91, 'y1': 73, 'y2': 93, 'page_number': 1}]
        new_q = {'title': title, 'weight': weight, 'crop_rect_list': new_crop}
        if parent_id:
            # TODO: This should throw a custom exception if a parent is not found
            parent = [parent for parent in new_q_data if parent['id'] == parent_id][0]
            if parent['children']:
                parent['children'].append(new_q)
            else:
                parent['children'] = [new_q]
        else:
            new_q_data.append(new_q)

        # TODO add id region support
        new_patch = {'assignment': {'identification_regions': {'name': None, 'sid': None}},
                     'question_data': new_q_data}

        outline_resp = self.course.session.get('https://www.gradescope.com/courses/' + self.course.cid +
                                               '/assignments/' + self.aid + '/outline/edit')
        parsed_outline_resp = BeautifulSoup(outline_resp.text, 'html.parser')
        authenticity_token = parsed_outline_resp.find('meta', attrs = {'name': 'csrf-token'} ).get('content')

        patch_resp = self.course.session.patch('https://www.gradescope.com/courses/' + self.course.cid +
                                               '/assignments/' + self.aid + '/outline/',
                                               headers = {'x-csrf-token': authenticity_token,
                                                          'Content-Type': 'application/json'},
                                               data = json.dumps(new_patch,separators=(',',':')))

        if patch_resp.status_code != requests.codes.ok:
            patch_resp.raise_for_status()

        # TODO this should be done smarter :(
        self.questions = []
        self._lazy_load_questions()

    # TODO allow this to be a predicate remove
    def remove_question(self, title=None, qid=None):
        if not title and not qid:
            return
        new_q_data = [q.to_patch() for q in self.questions]

        # TODO Yes this is slow and ugly, should be improved
        if title: 
            new_q_data = [q for q in new_q_data if q['title'] != title]
            for q in new_q_data:
                if q.get('children'):
                    q['children'] = [sq for sq in q['children'] if sq['title'] != title]
        else:
            new_q_data = [q for q in new_q_data if q['id'] != qid]
            for q in new_q_data:
                if q.get('children'):
                    q['children'] = [sq for sq in q['children'] if sq['id'] != qid]

        new_patch = {'assignment': {'identification_regions': {'name': None, 'sid': None}},
                     'question_data': new_q_data}

        outline_resp = self.course.session.get('https://www.gradescope.com/courses/' + self.course.cid +
                                               '/assignments/' + self.aid + '/outline/edit')
        parsed_outline_resp = BeautifulSoup(outline_resp.text, 'html.parser')
        authenticity_token = parsed_outline_resp.find('meta', attrs = {'name': 'csrf-token'} ).get('content')

        patch_resp = self.course.session.patch('https://www.gradescope.com/courses/' + self.course.cid +
                                               '/assignments/' + self.aid + '/outline/',
                                               headers = {'x-csrf-token': authenticity_token,
                                                          'Content-Type': 'application/json'},
                                               data = json.dumps(new_patch,separators=(',',':')))

        print(patch_resp.status_code)
    # TODO
    def add_instructor_submission(self, fname):
        '''
        Upload a PDF submission.
        '''
        pass

    # TODO
    def publish_grades(self):
        pass

    # TODO
    def unpublish_grades(self):
        pass

    def _lazy_load_questions(self):        
        outline_resp = self.course.session.get('https://www.gradescope.com/courses/' + self.course.cid +
                                               '/assignments/' + self.aid + '/outline/edit')
        parsed_outline_resp = BeautifulSoup(outline_resp.text, 'html.parser')

        props = parsed_outline_resp.find('div',
                                         attrs={'data-react-class':'AssignmentOutline'}).get('data-react-props')
        json_props = json.loads(props)
        outline = json_props['outline']

        for question in outline:
            qid = question['id']
            title = question['title']
            parent_id = question['parent_id']
            weight = question['weight']
            content = question['content']
            crop = question['crop_rect_list']
            children = []
            qchildren = question.get('children', [])
            
            for subquestion in qchildren:
                c_qid = subquestion['id']
                c_title = subquestion['title']
                c_parent_id = subquestion['parent_id']
                c_weight = subquestion['weight']
                c_content = subquestion['content']
                c_crop = subquestion['crop_rect_list']
                children.append(GSQuestion(c_qid, c_title, c_weight, [], c_parent_id, c_content, c_crop))
            self.questions.append(GSQuestion(qid, title, weight, children, parent_id, content, crop))
            
        



from enum import Enum
from bs4 import BeautifulSoup
from datetime import datetime
from gscdash.pyscope.person import GSPerson
from gscdash.pyscope.person import GSRole
from gscdash.pyscope.assignment import GSAssignment


class LoadedCapabilities(Enum):
    ASSIGNMENTS = 0
    ROSTER = 1


class GSCourse:
    def __init__(self, cid, name, shortname, year, session):
        """Create a course object that has lazy eval'd assignments"""
        self.cid = cid
        self.name = name
        self.shortname = shortname
        self.year = year
        self.session = session
        self.assignments = {}
        self.roster = {}  # TODO: Maybe shouldn't dict.
        self.state = (
            set()
        )  # Set of already loaded entities (TODO what is the pythonic way to do this?)

    # ~~~~~~~~~~~~~~~~~~~~~~PEOPLE~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def get_url(self):
        return "https://www.gradescope.com/courses/" + self.cid
    
    def get_roster(self):
        self._check_capabilities({LoadedCapabilities.ROSTER})

        return self.roster

    def add_person(self, name, email, role, sid=None, notify=False):
        self._check_capabilities({LoadedCapabilities.ROSTER})

        membership_resp = self.session.get(
            "https://www.gradescope.com/courses/" + self.cid + "/memberships"
        )
        parsed_membership_resp = BeautifulSoup(membership_resp.text, "html.parser")

        authenticity_token = parsed_membership_resp.find("meta", attrs={"name": "csrf-token"}).get(
            "content"
        )
        person_params = {
            "utf8": "✓",
            "user[name]": name,
            "user[email]": email,
            "user[sid]": "" if sid is None else sid,
            "course_membership[role]": role.value,
            "button": "",
        }
        if notify:
            person_params["notify_by_email"] = 1
        # Seriously. Why is this website so inconsistent as to where the csrf token goes?????????
        add_resp = self.session.post(
            "https://www.gradescope.com/courses/" + self.cid + "/memberships",
            data=person_params,
            headers={"x-csrf-token": authenticity_token},
        )

        # TODO this is highly wasteful, need to likely improve this.
        self.roster = {}
        self._lazy_load_roster()

    def remove_person(self, name):
        self._check_capabilities({LoadedCapabilities.ROSTER})

        membership_resp = self.session.get(
            "https://www.gradescope.com/courses/" + self.cid + "/memberships"
        )
        parsed_membership_resp = BeautifulSoup(membership_resp.text, "html.parser")

        authenticity_token = parsed_membership_resp.find("meta", attrs={"name": "csrf-token"}).get(
            "content"
        )
        remove_params = {"_method": "delete", "authenticity_token": authenticity_token}
        remove_resp = self.session.post(
            "https://www.gradescope.com/courses/"
            + self.cid
            + "/memberships/"
            + self.roster[name].data_id,
            data=remove_params,
            headers={"x-csrf-token": authenticity_token},
        )

        # TODO this is highly wasteful, need to likely improve this.
        self.roster = {}
        self._lazy_load_roster()

    def get_assignments(self):
        assignments_resp = self.session.get("https://www.gradescope.com/courses/" + self.cid + '/assignments')
        parsed = BeautifulSoup(assignments_resp.text, "html.parser")

        if 'You are not authorized to access this page.' in parsed.text:
            assignments_resp = self.session.get("https://www.gradescope.com/courses/" + self.cid)
            parsed = BeautifulSoup(assignments_resp.text, "html.parser")

        assignments_table = parsed.find("table", id="assignments-student-table")
        if not assignments_table:
            assignments_table = parsed.find("table", id="assignments-instructor-table")

        assignments_table_body = assignments_table.find("tbody")

        datefmt = "%Y-%m-%d %H:%M:%S %z"

        assignments = []

        for row in assignments_table_body.find_all("tr"):            
            name = row.find("th", class_="table--primaryLink")

            if not name:
                name = row.find("div", class_="assignments--rowTitle")
                assign_id = name.find('a').get('href').split('/')[-1]
            else:
                assign_id = name.find('a').get('href').split('/')[-1]
            name = name.text

            timeline_cols = row.find_all("td", {"class": "table--hiddenColumn"})

            if not timeline_cols or len(timeline_cols) == 0:
                timeline_cols = row.find_all("td", {"class": "hidden-column"})

            assigned = timeline_cols[0].text
            due = timeline_cols[1].text

            if assigned is None or assigned == "" or 'false' in assigned:
                assigned = None
            else:
                assigned = datetime.strptime(assigned, datefmt)

            if due is None or due == "":
                due = None
            else:
                due = datetime.strptime(due, datefmt)

            assignments.append( {
                    "id": assign_id,
                    "name": name,
                    "assigned": assigned,
                    "due": due,
                })

        return assignments

    def change_person_role(self, name, role):
        self._check_capabilities({LoadedCapabilities.ROSTER})

        membership_resp = self.session.get(
            "https://www.gradescope.com/courses/" + self.cid + "/memberships"
        )
        parsed_membership_resp = BeautifulSoup(membership_resp.text, "html.parser")

        authenticity_token = parsed_membership_resp.find("meta", attrs={"name": "csrf-token"}).get(
            "content"
        )
        role_params = {
            "course_membership[role]": role.value,
        }
        role_resp = self.session.patch(
            "https://www.gradescope.com/courses/"
            + self.cid
            + "/memberships/"
            + self.roster[name].data_id
            + "/update_role",
            data=role_params,
            headers={"x-csrf-token": authenticity_token},
        )

        # TODO this is highly wasteful, need to likely improve this.
        self.roster = {}
        self._lazy_load_roster()

    # ~~~~~~~~~~~~~~~~~~~~~~ASSIGNMENTS~~~~~~~~~~~~~~~~~~~~~~~~~~

    def add_assignment(
        self,
        name,
        release,
        due,
        template_file,
        student_submissions=True,
        late_submissions=False,
        group_submissions=0,
    ):
        self._check_capabilities({LoadedCapabilities.ASSIGNMENTS})

        assignment_resp = self.session.get(
            "https://www.gradescope.com/courses/" + self.cid + "/assignments"
        )
        parsed_assignment_resp = BeautifulSoup(assignment_resp.text, "html.parser")
        authenticity_token = parsed_assignment_resp.find("meta", attrs={"name": "csrf-token"}).get(
            "content"
        )

        # TODO Make this less brittle and make sure to support all options properly
        assignment_params = {
            "authenticity_token": authenticity_token,
            "assignment[title]": name,
            "assignment[student_submission]": student_submissions,
            "assignment[release_date_string]": release,
            "assignment[due_date_string]": due,
            "assignment[allow_late_submissions]": 1 if late_submissions else 0,
            "assignment[submission_type]": "image",  # TODO What controls this?
            "assignment[group_submission]": group_submissions,
        }
        assignment_files = {"template_pdf": open(template_file, "rb")}
        assignment_resp = self.session.post(
            "https://www.gradescope.com/courses/" + self.cid + "/assignments",
            files=assignment_files,
            data=assignment_params,
        )

        # TODO this is highly wasteful, need to likely improve this.
        self.assignments = {}
        self._lazy_load_assignments()

    def remove_assignment(self, name):
        self._check_capabilities({LoadedCapabilities.ASSIGNMENTS})

        assignment_resp = self.session.get(
            "https://www.gradescope.com/courses/"
            + self.cid
            + "/assignments/"
            + self.assignments[name].aid
            + "/edit"
        )
        parsed_assignment_resp = BeautifulSoup(assignment_resp.text, "html.parser")
        authenticity_token = parsed_assignment_resp.find("meta", attrs={"name": "csrf-token"}).get(
            "content"
        )

        remove_params = {"_method": "delete", "authenticity_token": authenticity_token}

        remove_resp = self.session.post(
            "https://www.gradescope.com/courses/"
            + self.cid
            + "/assignments/"
            + self.assignments[name].aid,
            data=remove_params,
        )

        # TODO this is highly wasteful, need to likely improve this.
        self.assignments = {}
        self._lazy_load_assignments()

    # ~~~~~~~~~~~~~~~~~~~~~~HOUSEKEEPING~~~~~~~~~~~~~~~~~~~~~~~~~

    def _lazy_load_assignments(self):
        """
        Load the assignment dictionary from assignments. This is done lazily to avoid slowdown caused by getting
        all the assignments for all classes. Also makes us less vulnerable to blocking.
        """
        assignment_resp = self.session.get(
            "https://www.gradescope.com/courses/" + self.cid + "/assignments"
        )
        parsed_assignment_resp = BeautifulSoup(assignment_resp.text, "html.parser")

        assignment_table = []
        for assignment_row in parsed_assignment_resp.findAll(
            "tr", class_="js-assignmentTableAssignmentRow"
        ):
            row = []
            for td in assignment_row.findAll("td"):
                row.append(td)
            assignment_table.append(row)

        for row in assignment_table:
            name = row[0].text
            aid = row[0].find("a").get("href").rsplit("/", 1)[1]
            points = row[1].text
            # TODO: (released,due) = parse(row[2])
            submissions = row[3].text
            percent_graded = row[4].text
            complete = True if "workflowCheck-complete" in row[5].get("class") else False
            regrades_on = False if row[6].text == "OFF" else True
            # TODO make these types reasonable
            self.assignments[name] = GSAssignment(
                name, aid, points, percent_graded, complete, regrades_on, self
            )
        self.state.add(LoadedCapabilities.ASSIGNMENTS)
        pass

    def _lazy_load_roster(self):
        """
        Load the roster list  This is done lazily to avoid slowdown caused by getting
        all the rosters for all classes. Also makes us less vulnerable to blocking.
        """
        membership_resp = self.session.get(
            "https://www.gradescope.com/courses/" + self.cid + "/memberships"
        )
        parsed_membership_resp = BeautifulSoup(membership_resp.text, "html.parser")

        if "You are not authorized to access this page" in membership_resp.text:
            print ('No access to roster, skipping')
            return
        else:
            print("Instructor access to roster")

        roster_table = []
        for student_row in parsed_membership_resp.find_all("tr", class_="rosterRow"):
            row = []
            for td in student_row("td"):
                row.append(td)
            roster_table.append(row)

        for row in roster_table:
            name = row[0].text.rsplit(" ", 1)[0]
            data_id = row[0].find("button", class_="rosterCell--editIcon").get("data-id")
            if len(row) >= 4 and row[2].find("option", selected="selected"):
                email = row[1].text
                role = row[2].find("option", selected="selected").text
                submissions = int(row[3].text)
                linked = True if "statusIcon-active" in row[4].find("i").get("class") else False
            else:
                email = row[2].text
                role = row[3].find("option", selected="selected").text
                if len(row[5].text):
                    submissions = int(row[5].text)
                    linked = True if "statusIcon-active" in row[6].find("i").get("class") else False
                else:
                    submissions = int(row[4].text)
                    linked = True if "statusIcon-active" in row[5].find("i").get("class") else False
            # TODO Make types reasonable.
            self.roster[name] = GSPerson(name, data_id, email, role, submissions, linked)
        self.state.add(LoadedCapabilities.ROSTER)

    def _check_capabilities(self, needed):
        """
        checks if we have the needed data loaded and gets them lazily.
        """
        missing = needed - self.state
        if LoadedCapabilities.ASSIGNMENTS in missing:
            self._lazy_load_assignments()
        if LoadedCapabilities.ROSTER in missing:
            self._lazy_load_roster()

    def delete(self):
        course_edit_resp = self.session.get(
            "https://www.gradescope.com/courses/" + self.cid + "/edit"
        )
        parsed_course_edit_resp = BeautifulSoup(course_edit_resp.text, "html.parser")

        authenticity_token = parsed_course_edit_resp.find("meta", attrs={"name": "csrf-token"}).get(
            "content"
        )

        print(authenticity_token)

        delete_params = {"_method": "delete", "authenticity_token": authenticity_token}
        print(delete_params)

        delete_resp = self.session.post(
            "https://www.gradescope.com/courses/" + self.cid,
            data=delete_params,
            headers={
                "referer": "https://www.gradescope.com/courses/" + self.cid + "/edit",
                "origin": "https://www.gradescope.com",
            },
        )

        # TODO make this less brittle



import requests
from bs4 import BeautifulSoup
from enum import Enum
from io import StringIO
from typing import List, Dict, Any
import logging
import pandas as pd
import lxml
from gscdash.pyscope.account import GSAccount
from gscdash.pyscope.course import GSCourse

from gscdash.course_info import CourseApi

class ConnState(Enum):
    INIT = 0
    LOGGED_IN = 1


class GSConnection(CourseApi):
    """The main connection class that keeps state about the current connection."""

    def __init__(self):
        """Initialize the session for the connection."""
        self.session = requests.Session()
        self.state = ConnState.INIT
        self.account = None

    def login(self, email, pswd, semesters):
        """
        Login to gradescope using email and password.
        Note that the future commands depend on account privilages.
        """
        init_resp = self.session.get("https://www.gradescope.com/")
        parsed_init_resp = BeautifulSoup(init_resp.text, "html.parser")
        for form in parsed_init_resp.find_all("form"):
            if form.get("action") == "/login":
                for inp in form.find_all("input"):
                    if inp.get("name") == "authenticity_token":
                        auth_token = inp.get("value")

        login_data = {
            "utf8": "✓",
            "session[email]": email,
            "session[password]": pswd,
            "session[remember_me]": 0,
            "commit": "Log In",
            "session[remember_me_sso]": 0,
            "authenticity_token": auth_token,
        }
        login_resp = self.session.post("https://www.gradescope.com/login", params=login_data)
        if len(login_resp.history) != 0:
            if login_resp.history[0].status_code == requests.codes.found:
                self.state = ConnState.LOGGED_IN
                self.account = GSAccount(email, self.session)
                self.get_account(semesters)
                return True
        else:
            return False

    def get_account(self, semesters):
        """
        Gets and parses account data after login. Note will return false if we are not in a logged in state, but
        this is subject to change.
        """
        if self.state != ConnState.LOGGED_IN:
            raise "No login"
            return False  # Should raise exception
        # Get account page and parse it using bs4
        account_resp = self.session.get("https://www.gradescope.com/account")
        parsed_account_resp = BeautifulSoup(account_resp.text, "html.parser")

        # Get instructor course data. This is called "Your Courses" if there is only
        # one type of course, or "Instructor Courses" if not
        instructor_courses = parsed_account_resp.find("h1", class_="pageHeading").next_sibling#.next_sibling

        if not instructor_courses.find_all("a", class_="courseBox"):
            instructor_courses = instructor_courses.next_sibling

        if not instructor_courses or len(instructor_courses) == 0:
            print ('No instructor courses found')

        for course in instructor_courses.find_all("a", class_="courseBox"):
            print ('Found course box')
            shortname = course.find("h3", class_="courseBox--shortname").text
            name = course.find("div", class_="courseBox--name").text
            cid = course.get("href").split("/")[-1]
            year = None
            print('Instructor in: ', cid, name, shortname)
            for tag in course.parent.previous_siblings:
                if "courseList--term" in tag.get("class"):
                    year = tag.string
                    break
            if year is None:
                raise "No year"
                return False  # Should probably raise an exception.
            
            if semesters is None or year in semesters:
                self.account.add_class(
                    cid, name, shortname, year, instructor=False
                )

        student_courses = parsed_account_resp.find(
            "h1", class_="pageHeading", string="Student Courses"
        )

        if student_courses is None:
            print('No student courses found')
            return True

        student_courses = student_courses.next_sibling

        for course in student_courses.find_all("a", class_="courseBox"):
            shortname = course.find("h3", class_="courseBox--shortname").text
            name = course.find("div", class_="courseBox--name").text
            cid = course.get("href").split("/")[-1]
            print('Student in: ', cid, name, shortname)

            for tag in course.parent.previous_siblings:
                if tag.get("class") == "courseList--term pageSubheading":
                    year = tag.body
                    break
            if year is None:
                raise "No year"
                return False  # Should probably raise an exception.
            if semesters is None or year in semesters:
                self.account.add_class(cid, name, shortname, year)

        return True

    def get_course_list(self) -> List[Dict]:
        icourses = []
        scourses = []
        if self.account.instructor_courses and len(self.account.instructor_courses):
            icourses = list(self.account.instructor_courses.values())
        if self.account.student_courses and len(self.account.student_courses):
            scourses = list(self.account.student_courses.values())

        return icourses + scourses
    
    def get_course_list_df(self) -> pd.DataFrame:
        return pd.DataFrame([vars(item) for item in self.get_course_list()]).drop(columns=['session','roster','assignments','state'])

    def get_course(self, cid) -> GSCourse:
        return self.get_course(cid)
    
    def get_assignments(self, course: GSCourse) -> List[Dict]:
        return course.get_assignments()
    
    # def get_assignments_df(self, course: GSCourse) -> pd.DataFrame:
    #     return super().get_assignments_df(course)
    

    def get_students(self, course: GSCourse) -> List[Dict]:
        return list(course.get_roster().values())
    
    def get_students_df(self, course: GSCourse) -> pd.DataFrame:
        ret = pd.DataFrame(self.get_students(course))
        ret.columns = ['sid', 'name', 'emails']
        return ret

    def get_assignment_submissions_df(self, course: GSCourse) -> pd.DataFrame:
        assignments = []
        for assignment in course.get_assignments():
            scores = self.session.get('https://gradescope.com/courses/' + course.cid + '/assignments/' + assignment['id'] + '/scores.csv').text

            print(scores)

            assignments.append(pd.read_csv(StringIO(scores)))
            assignments[-1]['course_id'] = course.cid
        return pd.concat(assignments)
    
    def get_assignment_submissions(self, course: GSCourse) -> List[Dict]:
         return self.get_assignment_submissions_df.to_list()
    
    def get_extensions_df_list(self, course: GSCourse) -> List[pd.DataFrame]:
        extensions = []
        for assignment in course.get_assignments():
            ext_table = self.session.get('https://gradescope.com/courses/' + course.cid + '/assignments/' + assignment['id'] + '/extensions').text

            # print("{}".format(ext_table))
            try:
                tab_list = pd.read_html(ext_table)

                print (tab_list)

                extensions.append(tab_list[0])
                extensions[-1]['course_id'] = course.cid
            except ValueError:
                pass
        return extensions
    
    def get_extensions_df(self, course: GSCourse) -> pd.DataFrame:
        return pd.concat(self.get_extensions_df_list(course))
    
    def get_extensions(self, course: GSCourse) -> List[Dict]:
        return self.get_extensions_df(course).to_list()



#####################################################################################################################
##
## Copyright (C) 2022-23 by Zachary G. Ives
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0.
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
##
#####################################################################################################################

from gscdash.pycanvas.pycanvas import CanvasConnection
from gscdash.course_info import CourseWrapper
import pandas as pd
from datetime import datetime
import pytz
import logging

from typing import Tuple, Any, List

class CanvasStatus(CourseWrapper):
    def __init__(self, canvas_url, canvas_key, filter_course_ids: List[str]):
        self.canvas = CanvasConnection(canvas_url, canvas_key)
        self.course_id_list = filter_course_ids

        
    def get_course_info(self) -> Tuple[Any]:
        canvas_courses = self.canvas.get_course_list_df()
        all_assignments = []
        all_students = []
        all_submissions = []
        all_student_summaries = []

        logging.debug('Getting course info from Canvas for {}'.format(self.course_id_list))

        rightnow = datetime.utcnow().replace(tzinfo=pytz.utc)
        for the_course in self.canvas.get_course_list_objs():
            print (the_course.name)

            if self.course_id_list and len(self.course_id_list) and the_course.id not in self.course_id_list:
                continue

            if (the_course.end_at and \
            (pd.to_datetime(the_course.start_at, utc=True) <= rightnow and rightnow <= pd.to_datetime(the_course.end_at, utc=True))):

                print ('{} through {}'.format(the_course.start_at, the_course.end_at))

                # OPTIONAL but not really needed since Quizzes are also Assignments?
                if False:
                    quizzes = self.canvas.get_quizzes_df(the_course)
                    if len(quizzes):
                        quizzes['course_id'] = the_course.id
                        print ('\nQuizzes:')
                        print(quizzes)

                # OPTIONAL: do we want modules and module items?
                if False:
                    modules = self.canvas.get_modules_df(the_course)
                    if len(modules):
                        modules['course_id'] = the_course.id
                        print ('\nModules:')
                        print(modules)
                    module_items = self.canvas.get_module_items_df(the_course)
                    if len(module_items):
                        print ('\nItems in modules:')
                        print(module_items)

                students = self.canvas.get_students_df(the_course)
                if len(students):
                    students['course_id'] = the_course.id
                    print ('\nStudents:')
                    print (students)
                    all_students.append(students)

                assignments = self.canvas.get_assignments_df(the_course)
                if len(assignments):
                    assignments['course_id'] = the_course.id
                    print ('\nAssignments:')
                    print(assignments)
                    all_assignments.append(assignments)

                # Get general student status, including late days etc
                student_summaries = self.canvas.get_student_summaries_df(the_course)
                if len(student_summaries):
                    print ('\nStudent summaries:')
                    print (student_summaries)
                    all_student_summaries.append(student_summaries)

                assignments = self.canvas.get_assignment_submissions_df(the_course)
                if len(assignments):
                    assignments['course_id'] = the_course.id
                    print ('\nAssignment submissions:')
                    print(assignments)
                    all_submissions.append(assignments)

        return ( canvas_courses,
                all_students,
                all_assignments,
                all_submissions,
                all_student_summaries)


#####################################################################################################################
##
## Copyright (C) 2022-23 by Zachary G. Ives
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0.
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
##
#####################################################################################################################

from gscdash.pyscope import pyscope as gs
from gscdash.pyscope.pyscope import GSConnection
from gscdash.pyscope.pyscope import GSAccount, GSCourse
from gscdash.course_info import CourseWrapper
import pandas as pd
from datetime import datetime
import pytz
import logging

from typing import Tuple, Any, List, Callable

class GradescopeStatus(CourseWrapper):
    def __init__(self, email: str, pwd: str, semesters: list):
        conn = gs.GSConnection()

        print("Logging into Gradescope as %s..."%email)
        if not conn.login(email, pwd, semesters):
            print ('Error: Failed to log in!')
            return None

        self.sem = semesters
        # Get Account
        success = conn.get_account(semesters)
        if not success:
            print("Error: Failed to get account!")
            return None

        self.gs = conn
        self.account = conn.account

    def get_courses(self) -> List[GSCourse]:
        """
        Returns a list of all courses for an individual -- both as a student and instructor
        """
        return list(self.account.student_courses.values()) + \
            list(self.account.instructor_courses.values())


    def get_semesters(self, courses: List[GSCourse]) -> list:
        semesters = set()
        for course in courses:
            semesters.add(course.year)
        return list(semesters)

    def get_course_events(self, course: GSCourse, all_events: list, event_handler: Callable[[GSCourse, dict], None]):
        course_name = course.shortname
        print(f"\nStarted Processing Course: {course_name}")

        assignments = course.get_assignments()
        # all_events = []

        for assign in assignments:
            print(assign['name'], 'is due', assign['due'])
            events = event_handler(course, assign)
            all_events.extend(events)

        print(f"Finished Processing Course: {course_name}")

        return


    def fetch_gradescope_events(self, event_handler: Callable[[GSCourse, dict], None], use_threads):
        courses = self.get_courses()

        # cal = Calendar()
        # threads = []
        events = []
        students = []

        # if use_threads:
        #     for course in courses:
        #         if self.sem is not None:
        #             if course.year != self.sem:
        #                 continue

        #         thread = Thread(target=get_course_events, args=(course, events))
        #         thread.start()
        #         threads.append(thread)

        #     for thread in threads:
        #         thread.join()

        #     for event in events:
        #         cal.events.add(event)
        # else:
        for course in courses:
            if self.sem is not None:
                if course.year != self.sem:
                    continue

            print('Processing course %s...'%course.name);
            
            self.get_course_events(course, events, event_handler)
            # for event in events:
            #     cal.events.add(event)

            print('Getting course roster')
            for row in course.get_roster():
                print(row)
                students.add(row)

        # return cal.serialize_iter()
        return events, students


        
    def get_course_info(self) -> Tuple[Any]:
        gs_courses = self.gs.get_course_list_df()
        all_assignments = []
        all_students = []
        all_submissions = []
        all_student_summaries = []

        logging.debug('Getting course info from Gradescope for {}'.format(self.sem))

        rightnow = datetime.utcnow().replace(tzinfo=pytz.utc)
        for the_course in self.gs.get_course_list():
            print (the_course.name)

            students = self.gs.get_students_df(the_course)
            if len(students):
                students['course_id'] = the_course.cid
                print ('\nStudents:')
                print (students)
                all_students.append(students)

            assignments = self.gs.get_assignments_df(the_course)
            if len(assignments):
                assignments['course_id'] = the_course.cid
                print ('\nAssignments:')
                print(assignments)
                all_assignments.append(assignments)

            assignments = self.gs.get_assignment_submissions_df(the_course)
            if len(assignments):
                assignments['course_id'] = the_course.cid
                print ('\nAssignment submissions:')
                print(assignments)
                all_submissions.append(assignments)

            extensions = self.gs.get_extensions_df(the_course)
            if len(extensions):
                print(extensions)
                all_student_summaries.append(extensions)

        return ( gs_courses,
                all_students,
                all_assignments,
                all_submissions,
                all_student_summaries)


#####################################################################################################################
##
## Copyright (C) 2022-23 by Zachary G. Ives
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0.
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
##
#####################################################################################################################

from abc import ABCMeta, abstractmethod
from typing import List, Tuple, Any
from datetime import datetime

import pandas as pd

class CourseDetails:
    course_uuid: str
    name: str
    starts: datetime
    ends: datetime

    def __str__(self, key):
        return '{}, {}, {}, {}'.format(self.course_uuid, self.name, self.starts, self.ends)
    
    def __iter__(self):
        yield self.course_uuid
        yield self.name
        yield self.starts
        yield self.ends 

class Assignment:
    assignment_uuid: str
    name: str
    due: datetime
    late: datetime
    points: int
    ec_points: int
    weight: float

class Person:
    student_id: str
    name: str
    emails: List[str]

    def __str__(self, key):
        return '{}, {}, {}'.format(self.student_id, self.name, self.emails)
    
    def __iter__(self):
        yield self.student_id
        yield self.name
        yield self.emails

class CourseApi:
    @abstractmethod
    def get_course_list(self) -> List[CourseDetails]:
         raise NotImplementedError()
    
    def get_course_list_df(self) -> pd.DataFrame:
        return pd.DataFrame(self.get_course_list())

    @abstractmethod    
    def get_assignments(self, course) -> List[Assignment]:
         raise NotImplementedError()
    
    def get_assignments_df(self, course) -> pd.DataFrame:
        return pd.DataFrame(self.get_assignments(course))

    @abstractmethod
    def get_quizzes(self, course) -> List[Assignment]:
        pass

    def get_quizzes_df(self, course) -> pd.DataFrame:
        return self.get_quizzes(course)

    @abstractmethod
    def get_students(self, course) -> List[Person]:
        raise NotImplementedError()
 
    def get_students_df(self, course) -> pd.DataFrame:
        return pd.DataFrame(self.get_students(course))
    
    @abstractmethod
    def get_assignment_submissions(self, course) -> List[Assignment]:
        raise NotImplementedError()
    
    def get_assignment_submissions_df(self, course) -> pd.DataFrame:
        return pd.DataFrame(self.get_assignment_submissions(course))

 
    # @abstractmethod
    # def get_instructors(self, course) -> List[Person]:
    #     raise NotImplementedError()

    # def get_instructors_df(self) -> pd.DataFrame:
    #     return pd.DataFrame(self.get_instructors())

class CourseWrapper(object):
    @abstractmethod
    def get_course_info(self) -> Tuple[Any]:
        """ 
        Fetch a tuple of lists of dataframes for each course, including students, assignments, submissions.
        """
        raise NotImplementedError()




from gscdash.pyscope import pyscope as gs
from ics import Calendar, Event
import datetime
import os
import hashlib

from pyscope.account import GSAccount, GSCourse


def create_hw_events(course: GSCourse, assignment: dict) -> list:
    name = assignment["name"]
    # print(f"\tProcessing Assignment: {name}")

    assigned = assignment["assigned"]
    due = assignment["due"]

    date_printfmt = "%A, %d. %B %Y %I:%M%p"

    desc = f"""Course: {course.name}
Course Shortname: {course.shortname}
Link to Course: {course.get_url()}

Assignment: {name}
Assigned: {assigned.strftime(date_printfmt) if assigned is not None else "None"}
Due: {due.strftime(date_printfmt) if due is not None else "None"}"""

    out = []

    def hw_event(name:str, course_name:str, type, time):
        event = Event()
        event.name = f"{name} - {course_name} {type}"
        event.begin = time
        event.end = time
        event.description = desc

        hash = f"{name}{course_name}{time.strftime('%Y%m%d')}"
        hash = hashlib.sha1(hash.encode("utf-8")).hexdigest()
        event.uid = f"{hash}@cis.upenn.edu"
        return event

    if assigned is not None:
        out.append(hw_event(name, course.shortname, "Assigned", assigned))

    if due is not None:
        out.append(hw_event(name, course.shortname, "Due", due))

    return out





from bs4 import BeautifulSoup

class GSQuestion():

    def __init__(self, qid, title, weight, children, parent_id, content, crop):
        '''Create a assignment object'''
        self.title = title
        self.qid = qid
        self.children = children
        self.weight = weight
        self.parent_id = parent_id
        self.content = content
        self.crop = crop
        
    def to_patch(self):
        children = [child.to_patch() for child in self.children]
        output = {'id': self.qid, 'title': self.title, 'weight': self.weight, 'crop_rect_list': self.crop}
        print('length of children:', len(self.children))
        if len(children) != 0:
            output['children'] = children
        return output



import setuptools

with open("README.md", "r", encoding = "utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name = "gscdash",
    version = "0.0.4",
    author = "Zack Ives",
    author_email = "zives@cis.upenn.edu",
    description = "Gradescope-Canvas Dashboard support",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/upenn/gradescope-canvas-dashboard",
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "requests",
        "html5lib",
        "beautifulsoup4",
        "canvasapi",
        "pandas",
        "pytz",
        "pyyaml",
    ],
    packages = ["gscdash", "gscdash.pycanvas", "gscdash.pyscope"],
    python_requires = ">=3.9"
)


#####################################################################################################################
##
## Copyright (C) 2022-23 by Zachary G. Ives
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0.
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
##
#####################################################################################################################

from canvasapi import Canvas
from canvasapi.course import Course
from canvasapi.module import Module
from canvasapi.assignment import Assignment
from canvasapi.exceptions import *
from canvasapi.paginated_list import PaginatedList
import pandas as pd
from gscdash.course_info import CourseApi
import logging
from typing import List, Dict

class CanvasConnection(CourseApi):
    def __init__(self, canvas_url, canvas_key):
        self.canvas = Canvas(canvas_url, canvas_key)

        self.get_course_list_full()
        return
    
    def _get_paginated(the_list, paginated: PaginatedList):
        for item in paginated:
            the_list.append(item)

        return the_list
    
    def get_course_list(self) -> List[Dict]:
        return self.courses

    def get_course_list_objs(self) -> List[Course]:
        return self.course_objs
    
    def get_course_list_df(self) -> pd.DataFrame:
        return pd.DataFrame(self.get_course_list())
    
    def get_course_list_full(self) -> pd.DataFrame:
        course_list = []
        CanvasConnection._get_paginated(course_list, self.canvas.get_courses(per_page=100))
        self.courses = []
        self.course_objs = []
        for course in course_list:
            self.course_objs.append(course)
            self.courses.append({
                'id': course.id,
                'name': course.name,
                'start_at': course.start_at,
                'end_at': course.end_at,
                'workflow_state': course.workflow_state,
                'course_code': course.course_code,
                # 'sis_course_id': course.sis_course_id,
                # 'integration_id': course.integration_id,
                # 'hide_final_grades': course.hide_final_grades,
                'is_public': course.is_public,
           })
            
        return self.courses
    
    def get_course(self, course_id: str) -> Course:
        return self.canvas.get_course(course_id)

    def get_quizzes(self, course: Course) -> List[Dict]:
        """
        Returns a list of dictionaries containing the quizzes for a course.
        """
        quizzes = []
        try:
            quiz_list = []
            CanvasConnection._get_paginated(quiz_list, course.get_quizzes(per_page=100))
            for quiz in quiz_list:
                quizzes.append({'id':quiz.id,'title':quiz.title,'published':quiz.published,\
                    'unlock_at': quiz.unlock_at, 'due_at': quiz.due_at, 'lock_at': quiz.lock_at, 'published': quiz.published})
        except ResourceDoesNotExist:
            logging.warning('No quizzes found for course %s', course)
            pass

        return quizzes
    
    def get_quizzes_df(self, course: Course) -> pd.DataFrame:
        return pd.DataFrame(self.get_quizzes(course))

    def get_modules(self, course: Course) -> List[Dict]:
        """
        Returns a list of dictionaries containing the modules for a course.
        """
        modules = []
        mods = []
        CanvasConnection._get_paginated(mods, course.get_modules(per_page=100))
        for module in mods:
            try:
                modules.append({
                    'id': module.id,
                    'name': module.name,
                    'published': module.published,
                    'unlock_at': module.unlock_at
                })
            except AttributeError:
                modules.append({
                    'id': module.id,
                    'name': module.name,
                    'unlock_at': module.unlock_at
                })

        return modules
    
    def get_modules_df(self, course: Course) -> pd.DataFrame:
        return pd.DataFrame(self.get_modules(course))

    def get_module_items(self, course: Course) -> List[Dict]:
        """
        Returns a list of dictionaries containing the module items for a course.
        """
        module_items = []
        mods = []
        CanvasConnection._get_paginated(mods, course.get_modules(per_page=100))
        for module in mods:
            mod_items = []
            CanvasConnection._get_paginated(mod_items, module.get_module_items(per_page=100))
            for item in mod_items:
                try:
                    details = {
                        'module_id': module.id,
                        'module_name': module.name,
                        'id': item.id,
                        'title': item.title,
                        # 'published': item.published,
                        'type': item.type,
                        'html_url': item.html_url
                    }
                except AttributeError:
                    details = {
                        'module_id': module.id,
                        'module_name': module.name,
                        'id': item.id,
                        'title': item.title,
                        # 'published': item.published,
                        'type': item.type
                    }

                if item.type == 'Quiz':
                    details['url'] = item.url
                if item.type == 'ExternalUrl':
                    details['external_url'] = item.external_url

                try:
                    details['published'] = item['published']
                except:
                    pass
                module_items.append(details)
        return module_items

    def get_module_items_df(self, course: Course) -> pd.DataFrame:
        return pd.DataFrame(self.get_module_items(course))

    def get_matching_module_url(self, module_items: List[Module], index: str, typ: str) -> str:
        matches = module_items[module_items['title'].apply(lambda x: x.startswith(index))]
        if len(matches):
            matches = matches[matches['type'] == typ]
        if len(matches):
            if typ == 'Quiz':
                return matches['html_url'].array[0]
            else:
                return matches['external_url'].array[0]
            
    def get_assignments(self, course: Course) -> List[Dict]:
        """
        Returns a list of dictionaries containing the assignments for a course.
        """
        assignments = []
        ret = []
        CanvasConnection._get_paginated(assignments, course.get_assignments(per_page=100))
        for assignment in assignments:
            ret.append(vars(assignment))
            del(ret[-1]['_requester'])
            del(ret[-1]['description'])
            del(ret[-1]['secure_params'])
        return ret
            
    def get_student_summaries_df(self, course: Course) -> pd.DataFrame:
        """
        Returns a dataframe containing the student summaries for a course.
        """
        try:
            the_list = []
            summaries = []
            CanvasConnection._get_paginated(the_list, course.get_course_level_student_summary_data(per_page=100))
            for item in the_list:
                summaries.append(vars(item))
                del(summaries[-1]['_requester'])
            ret = pd.DataFrame(summaries)
            if 'tardiness_breakdown' in ret.columns:
                return ret.join(pd.json_normalize(ret['tardiness_breakdown'])).drop(['tardiness_breakdown'], axis=1)
            else:
                return ret
        except Forbidden:
            return pd.DataFrame()
        except ResourceDoesNotExist:
            return pd.DataFrame()

            

    def get_students(self, course) -> List[Dict]:
        students = []
        ret= []
        try:
            CanvasConnection._get_paginated(students, course.get_users(enrollment_type=['student'], per_page=100))
            for student in students:
                ret.append(vars(student))
                del(ret[-1]['_requester'])
        except ResourceDoesNotExist:
            logging.warning('No students found for course %s', course)
            pass
        except Forbidden:
            logging.warning('Unauthorized to access students for course %s', course)
            pass
    
        return ret
    
    def get_assignment_submissions(self, course: Course) -> List[Dict]:
        """
        Returns a list of dictionaries containing the assignment submissions for a course.
        """
        ret = []
        assignments = []
        CanvasConnection._get_paginated(assignments, course.get_assignments(per_page=100))
        try:
            for assignment in assignments:
                submissions = []
                CanvasConnection._get_paginated(submissions, assignment.get_submissions(per_page=100))
                for submission in submissions:
                    ret.append(vars(submission))
                    ret[-1]['assignment_id'] = assignment.id
                    del(ret[-1]['_requester'])
        except ResourceDoesNotExist:
            pass
        except Forbidden:
            pass
        return ret
    
    def get_assignment_submissions_df(self, course: Course) -> pd.DataFrame:
        return pd.DataFrame(self.get_assignment_submissions(course))


from enum import Enum
from gscdash.course_info import Person

class GSRole(Enum):
    STUDENT = 0
    INSTRUCTOR = 1
    TA = 2
    READER = 3
    
    def from_str(val):
        if isinstance(val, GSRole):
            return val
        strings = {
            'Instructor': GSRole.INSTRUCTOR,
            'Student': GSRole.STUDENT,
            'TA': GSRole.TA,
            'Reader': GSRole.READER
        }
        role =  strings.get(val)
        if role is not None:
            return role
        else:
            raise GSRoleException("Not a valid role string: " + role)  

    def to_str(val):
        strings = {
            GSRole.INSTRUCTOR : 'Instructor',
            GSRole.STUDENT : 'Student',
            GSRole.TA : 'TA',
            GSRole.READER : 'Reader'
        }
        return strings[val]
        
    class GSRoleException(Exception):
        pass
    
    
class GSPerson(Person):
    def __init__(self, name, data_id, email, role, submissions, linked):
        self.name = name
        self.data_id = data_id
        self.student_id = data_id
        self.email = email
        self.emails = [email]
        self.role = GSRole.from_str(role)
        self.linked = linked
        self.submissions = submissions



from bs4 import BeautifulSoup
from gscdash.pyscope.course import GSCourse

class GSAccount():
    '''A class designed to track the account details (instructor and student courses)'''

    def __init__(self, email, session):
        self.email = email
        self.session = session
        self.instructor_courses = {}
        self.student_courses = {}

    def add_class(self, cid, name, shortname, year, instructor = False):
        if instructor:
            self.instructor_courses[cid] = GSCourse(cid, name, shortname, year, self.session)
        else:
            self.student_courses[cid] = GSCourse(cid, name, shortname, year, self.session)

    # TODO add default exceptions when doing unsafe things.
    def delete_class(self, cid):
        self.instructor_courses[cid].delete()
        del self.instructor_courses[cid]

    def create_course(self, name, shortname, description, term, year, school, entry_code_enabled = False):
        '''Returns course ID'''
        account_resp = self.session.get("https://www.gradescope.com/account")
        parsed_account_resp = BeautifulSoup(account_resp.text, 'html.parser')

        create_modal = parsed_account_resp.find('div', id = 'createCourseModal')
        authenticity_token = create_modal.find('input', attrs = {'name': 'authenticity_token'}).get('value')
        schools = create_modal.find('select', id = 'course_school_id')
        school_id = schools.find('option', text = school).get('value') # TODO Fix this on bad params.
        course_data = {
            "utf8": "✓",
            "authenticity_token": authenticity_token,
            "course[shortname]": shortname,
            "course[name]": name,
            "course[description]": description,
            "course[term]": term,
            "course[year]": year,
            "course[school_id]": school_id,
            "course[entry_code_enabled]": 1 if entry_code_enabled else 0,
            "commit": "Create Course",
        }
        
        course_resp = self.session.post("https://www.gradescope.com/courses", params=course_data)
        # TODO This is brittle
        cid = course_resp.history[0].headers.get('Location').rsplit('/', 1)[1]
        # TODO fix term, year union
        self.add_class(cid, name, shortname, term+" "+year, instructor = True )
        return cid
        
